/*
 * 9.2. Clasa completa arbore binar de cautare. Fisierul header.
 *
 *      Au fost introduse procedurile de salvare/restaurare din
 *      Sectiunea 9.5. 
 */

#ifndef __ARBORE_H
#define __ARBORE_H

#include "tablou.h"


template <class E>
class varf {
  friend class arbore<E>;
  friend class s8a<E>;

  friend varf<E>* _search( varf<E>*, const E& );
  friend varf<E>* _min   ( varf<E>*           );
  friend varf<E>* _succ  ( varf<E>*           );
  friend varf<E>* _copy  ( varf<E>*           );

  friend void      _inord( varf<E>* );

private:
  varf( const E& v, float f = 0 ): key( v )
    { st = dr = tata = 0; p = f; }

  ~varf( ) { delete st; delete dr; }

  varf<E>   *st; // adresa fiului stang
  varf<E>   *dr; // adresa fiului drept
  varf<E> *tata; // adresa varfului tata

  E         key; // cheia
  float       p; // frecventa utilizarii cheii curente
};


template <class E>
varf<E>* _search( varf<E>* x, const E& k ) {
  while ( x != 0 && k != x->key )
    x = k > x->key? x->dr: x->st;
  return x;
}

template <class E>
varf<E>* _min( varf<E>* x ) {
  while ( x->st != 0 )
    x = x->st;
  return x;
}

template <class E>
varf<E>* _succ( varf<E>* x ) {
  if ( x->dr != 0 ) return _min( x->dr );

  varf<E> *y = x->tata;
  while ( y != 0 && x == y->dr )
    { x = y; y = y->tata; }
  return y;
}

template <class E>
varf<E>* _copy( varf<E>* x ) {
  varf<E> *z = 0;
  if ( x ) {
    // varful nou alocat se initializeaza cu x
    z = new varf<E>( x->key, x->p );

    // se copiaza subarborii din stanga si din deapta; in
    // fiecare se initializeaza legatura spre varful tata
    if ( (z->st = _copy( x->st )) != 0 )
      z->st->tata = z;
    if ( (z->dr = _copy( x->dr )) != 0 )
      z->dr->tata = z;
  }
  return z;
}

template <class E>
void _inord( varf<E> *x ) {
  if ( !x ) return;

  _inord( x->st );
  cout << x
       << " ( key " << x->key
       << ", f "    << x->p
       << ", st "   << x->st
       << ", dr "   << x->dr
       << ", tata " << x->tata
       << " )\n";
  _inord( x->dr );
}



template <class E>
class s8a {   // clasa pentru construirea arborelui optim
  friend class arbore<E>;
private:
  s8a( varf<E> *root, int nn ): pvarf( n = nn ) {
    int i = 0;          // indice in pvarf
    setvarf( i, root ); // setarea elementelor din pvarf
  }

  // initializarea tabloului pvarf cu un arbore deja format
  void setvarf( int&, varf<E>* );

  varf<E>* greedy( ) {  // "optim" prin algoritmul greedy
    return _greedy( 0, n );
  }

  varf<E>* prodin( ) {  // optim prin programare dinamica
    _progDinInit( ); return _progDin( 0, n - 1 );
  }

  // functiile prin care se formeaza efectiv arborele
  varf<E>* _greedy     ( int, int );
  varf<E>* _progDin    ( int, int );
  void     _progDinInit( );   // initializeaza tabloul r

  // date membre
  tablou<varf<E>*> pvarf;  // tabloul adreselor varfurilor
  int n;                   // numarul varfurilor din arbore

  // tabloul indicilor necesar alg. de programare dinamica
  tablou< tablou<int> > r;
};

template <class E>
void s8a<E>::setvarf( int& poz, varf<E>* x ) {
  if ( x ) {
    setvarf( poz, x->st );
    pvarf[ poz++ ] = x;
    setvarf( poz, x->dr );

    // anulam toate legaturile elementului x
    x->st = x->dr = x->tata = 0;
  }
}

template <class E>
varf<E>* s8a<E>::_greedy( int m, int M ) {
  // m si M sunt limitele subsecventei curente
  if ( m == M ) return 0;

  // se determina pozitia k a celei mai frecvente chei
  int k;                  float pmax = pvarf[ k = m ]->p;
  for ( int i = m; ++i < M; )
    if ( pvarf[ i ]->p > pmax ) pmax = pvarf[ k = i ]->p;

  // se selecteaza adresa varfului de pe pozitia k
  varf<E> *actual = pvarf[ k ];

  // se construiesc subarborii din stanga si din deapta
  // se initializeaza legatura spre varful tata
  if ( (actual->st = _greedy( m,     k )) != 0 )
    actual->st->tata = actual;
  if ( (actual->dr = _greedy( k + 1, M )) != 0 )
    actual->dr->tata = actual;

  // subarborele curent este gata; se returneaza adresa lui
  return actual;
}

template <class E>
varf<E>* s8a<E>::_progDin( int i, int j ) {
  // i si j, i <=j, sunt coordonatele radacinii
  //  subarborelui curent in tabloul r
  if ( i > j ) return 0;

  // se selecteaza adresa varfului radacina
  varf<E> *actual = pvarf[ r[ j ][ i ] ];

  if ( i != j ) { // daca nu este un varf frunza ...
    // se construiesc subarborii din stanga si din deapta
    // se initializeaza legatura spre varful tata
    if ( (actual->st = _progDin( i, r[j][i] - 1 )) != 0 )
       actual->st->tata = actual;
    if ( (actual->dr = _progDin( r[j][i] + 1, j )) != 0 )
       actual->dr->tata = actual;
  }

  // subarborele curent este gata; se returneaza adresa lui
  return actual;
}

template <class E>
void s8a<E>::_progDinInit( ) {
  int i, j, d;
 tablou< tablou<float> > C;   // tabloul C este local

  // redimensionarea si initializarea tablourilor C si r
  //  ATENTIE! tablourile C si r sunt TRANSPUSE.
  r.newsize( n );
  C.newsize( n );
  for ( i = 0; i < n; i++ ) {
    r[ i ].newsize( i + 1 ); r[ i ][ i ] = i;
    C[ i ].newsize( i + 1 ); C[ i ][ i ] = pvarf[ i ]->p;
  }

  // pentru inceput C este identic cu m
  for ( d = 1; d < n; d++ )
    for ( i = 0; (j = i + d) < n; i++ )
      C[ j ][ i ] = C[ j - 1 ][ i ] + C[ j ][ j ];

  // elementele din C se calculeaza pe diagonale
  for ( d = 1; d < n; d++ )
    for ( i = 0; (j = i + d) < n; i++ ) {
      // in calculul minimului dintre C[i][k-1]+C[k+1][j]
      //  consideram mai intai cazurile k=i si k=j in care
      //  avem C[i][i-1] = 0 si C[j+1][j] = 0
      int k; float Cmin;
      if ( C[ j ][ i + 1 ] < C[ j - 1 ][ i ] )
        Cmin = C[ j ][ (k = i) + 1 ];
      else
        Cmin = C[ (k = j) - 1 ][ i ];

      // au mai ramas de testat elementele i+1, ..., j-1
      for ( int l = i + 1; l < j; l++ )
        if ( C[ l - 1 ][ i ] + C[ j ][ l + 1 ] < Cmin )
          Cmin = C[ (k = l) - 1 ][ i ] + C[ j ][ l + 1 ];

      // minimul si pozitia lui sunt stabilite ...
      C[ j ][ i ] += Cmin;
      r[ j ][ i ] = k;
    }
}



template <class E>
class arbore {
  friend class s8a<E>;
public:
  arbore( )  { root = 0; n = 0; }
  arbore( const arbore<E>& );

  ~arbore( ) { delete root; }
  arbore& operator =( const arbore& );

  int ins   ( const E&, float = 0 );
  int del   (       E&            );
  int search(       E&            );

  arbore& re_greedy( );
  arbore& re_prodin( );

  int save( char* );
  int rest( char* );

  void inord( ) { _inord( root ); }
private:
  varf<E> *root;  // adresa varfului radacina
  int      n;     // numarul varfurilor din arbore
};

template <class E>
arbore<E>::arbore( const arbore<E>& a ) {
  root = _copy( a.root ); n = a.n;
}

template <class E>
arbore<E>& arbore<E>::operator =( const arbore<E>& a ) {
  delete root;
  root = _copy( a.root ); n = a.n;
  return *this;
}

template <class E>
arbore<E>& arbore<E>::re_greedy( ) {
// reorganizare prin metoda greedy
  s8a<E> opt( root, n );
  root = opt.greedy( );

  return *this;
}

template <class E>
arbore<E>& arbore<E>::re_prodin( ) {
// reorganziare prin programare dinamica
  s8a<E> opt( root, n );
  root = opt.prodin( );

  return *this;
}

template <class E>
int arbore<E>::search( E& k ) {
  varf<E> *x = _search( root, k );
  if ( !x )   return 0; // element absent
  x->p++;               // actualizarea frecventei
  k = x->key; return 1;
}

template <class E>
int arbore<E>::ins( const E& k, float p ) {
  varf<E> *y = 0, *x = root;

  while ( x != 0 ) {
    y = x;
    if ( k == x->key ) { // cheia deja exista in arbore
      x->p += p;         // se actualizeaza frecventa
      return 0;          // se returneaza cod de eroare
    }
    x = k > x->key? x->dr: x->st;
  }

  // cheia nu exista in arbore
  varf<E> *z = new varf<E>( k, p );
  z->tata = y;

  if ( y == 0 )                  root = z;
  else   if ( z->key > y->key ) y->dr = z;
         else                   y->st = z;

  n++;   // in arbore este cu un varf mai mult
  return 1;
}

template <class E>
int arbore<E>::del( E& k ) {
  varf<E> *z = _search( root, k );  // se cauta cheia k
  if ( !z ) return 0;               // nu a fost gasita

  n--;        // in arbore va fi cu un varf mai putin
  k = z->key; // k va retine intreaga informatie din z

  //  - y este z daca z are cel mult un fiu si
  //      succesorul lui z daca z are doi fii
  //  - x este fiul lui y sau 0 daca y nu are fii
  varf<E> *y, *x;

  y = z->st == 0 || z->dr == 0? z: _succ( z );
  x = y->st != 0?           y->st: y->dr;

  // se elimina varful y din arbore astfel:
  //  1. se stabileste legatura in x spre varful tata
  if ( x != 0 )
    x->tata = y->tata;

  //  2. in varful tata se stabileste legatura spre x
  if ( y->tata == 0 )                   root = x;
  else   if ( y == y->tata->st ) y->tata->st = x;
         else                    y->tata->dr = x;

  //  3. daca z are 2 fii, succesorul lui ii ia locul
  if ( y != z ) { z->key = y->key; z->p = y->p; }

  //  4. stergerea propriu-zisa
  y->st = y->dr = 0;
  delete y;

  return 1;
}


#include <fstream.h>
#include "coada.h"

template <class E>
int arbore<E>::save( char *file ) {
  ofstream f( file, ios::binary );  // deschide fisierul
  if ( !f ) return 0;               // eroare la deschidere

  coada<varf<E>*> c( n + 1 ); // ptr. parcurgerea in latime
  varf<E> *x;                 // varful curent

  c.ins_q( root );            // primul element din coada
  while ( c.del_q( x ) ) {
    if ( !f.write( (char *) &(x->key), sizeof( x->key ) ) )
      return 0;               // eroare la scriere
    if ( !f.write( (char *) &(x->p  ), sizeof( x->p   ) ) )
      return 0;               // eroare la scriere

    if ( x->st ) c.ins_q( x->st );
    if ( x->dr ) c.ins_q( x->dr );
  }
  f.close( );
  return 1;
}

template <class E>
int arbore<E>::rest( char *file ) {
  ifstream f( file, ios::binary ); // deschide fisierul
  if ( !f ) return 0;              // eroare la deschidere

  delete root;
  root = 0; n = 0; // se va crea un nou arbore

  E key; float p;  // informatia din varful curent
  while ( f.read( (char *) &key, sizeof( key ) ) &&
          f.read( (char *) &p,   sizeof( p   ) )    )
    ins( key, p );

  f.close( );
  return 1;
}

#endif
