/*
 * 4.3.1. Clasa parametrica lista<T>. Fisierul header.
 *
 *  S-a adaugat:
 *     -  clasa lmsort (sortarea prin interclasare) din Sectiunea 7.4.2.
 *
 */

#ifndef __LISTA_H
#define __LISTA_H

#include <iostream.h>

template <class E>
class nod {
  friend class lista<E>;
  friend class iterator<E>;
  friend class lmsort<E>;

protected:
  nod( const E& v ): val( v ) { next = 0; }

  E       val;  // informatia propriu-zisa
  nod<E> *next; // adresa nodului urmator
};

template <class E>
class lista {
  friend class iterator<E>;
  friend class lmsort<E>;

public:
  lista(  ) { head = 0; }
  lista( const lista<E>& s ) { init( s ); }
  ~lista( ) { reset( ); }

  lista& operator =( const lista<E>& );
  lista& insert( const E& );

private:
  nod<E> *head;  // adresa primul nod din lista

  void init( const lista<E>& );
  void reset( );
};

template <class E>
lista<E>& lista<E>::insert( const E& v ) {
  nod<E> *pn = new nod<E>( v );
  pn->next = head;
  head = pn;

  return *this;
}

template <class E>
lista<E>& lista<E>::operator =( const lista<E>& s ) {
  reset( );
  init( s );

  return *this;
}

template <class E>
void lista<E>::init( const lista<E>& sursa ) {
  E v; iterator<E> s = sursa;

  for ( nod<E> *tail = head = 0; s( v ); ) {
    nod<E> *pn = new nod<E>( v );
    if ( !tail ) head = pn; else  tail->next = pn;
    tail = pn;
  }
}

template <class E>
void lista<E>::reset( ) {
  nod<E> *a = head;

  while( a ) {
    nod<E> *pn = a->next;
    delete a;  a = pn;
  }
  head = 0;
}

template <class E>
ostream& operator <<( ostream& os, const lista<E>& lista ) {
  E v; iterator<E> l = lista;

  os << " { ";
  while ( l( v ) ) os << v << ' ';
  os << "} ";

  return os;
}

/*
 * Aceasta procedura este necesara datorita unor
 * anomalii din bibliotecile de I/E Borland 3.1 si 4.0.
 */
template <class E>
ostream& operator <<( ostream_withassign& os, lista<E>& l ) {
  return ( os << dec << l );
}

template <class E>
class iterator {
public:
  iterator( );
  iterator( const lista<E>& );

  operator ()( E& );
  iterator<E>& operator =( const lista<E>& );

private:
  nod<E>* const *phead;
  nod<E>        *a;
};

template <class E>
iterator<E>::iterator( ) {
  phead = 0;
  a = 0;
}

template <class E>
iterator<E>::iterator( const lista<E>& l ) {
  phead = &l.head;
  a = *phead;
}

template <class E>
iterator<E>::operator ()( E& v ) {
  if( a ) { v = a->val;  a = a->next; return 1; }
  else    { if( *phead ) a = *phead;  return 0; }
}

template <class E>
iterator<E>& iterator<E>::operator =( const lista<E>& l ) {
  phead = &l.head;
  a = *phead;

  return *this;
}



/*
 * sortare prin interclasare (mergesort)
 */

template <class E>
lista<E>& mergesort( lista<E>& l ) {
  ( lmsort<E> )l;
  return l;
}


// clasa pentru implementarea sortarii prin interclasare
template <class E>
class lmsort {
public:
  lmsort( lista<E>& );

private:
  nod<E>* mergesort( nod<E>* );
  nod<E>* merge( nod<E>*, nod<E>* );
};

template <class E>
lmsort<E>::lmsort( lista<E>& l ) {
  if ( l.head )
    l.head = mergesort( l.head );
}

template <class E>
nod<E>* lmsort<E>::mergesort ( nod<E> *c ) {
  if ( c && c->next ) {
    // sunt cel putin doua noduri in lista
    nod<E> *a = c, *b;

    for ( b = c->next; b; a = a->next )
      if ( b->next ) b = b->next->next;
      else           break;
    b = a->next; a->next = 0;
    return merge( mergesort( c ), mergesort( b ) );
  }
  else
    // lista contine cel mult un nod
    return c;
}

template <class E>
nod<E>* lmsort<E>::merge( nod<E> *a, nod<E> *b ) {
  nod<E> *head;           // primul nod al listei interclasate

  if ( a && b )
    // ambele liste sunt nevide;
    // stabilim primul nod din lista interclasata
    if ( a->val > b->val ) { head = b; b = b->next; }
    else                   { head = a; a = a->next; }
  else
    // cel putin una din liste este vida;
    // nu avem ce interclasa
    return a? a: b;

  // interclasarea propriu-zisa
  nod<E> *c = head;       // ultimul nod din lista interclasata
  while ( a && b )
    if ( a->val > b->val ) { c->next = b; c = b; b = b->next; }
    else                   { c->next = a; c = a; a = a->next; }

  // cel putin una din liste s-a epuizat
  c->next = a? a: b;

  // se returneaza primul nod al listei interclasate
  return head;
}

#endif