/*
 * 4.1.3. Clasa parametrica tablou<T>. Fisierul header.
 *
 *   Contine completarile sugerate in Sectiunea 4.2.2 pentru
 *   manipularea obiectelor constante:
 *     - operatorul size este de tip const
 *     - s-a introdus un nou operator [] de tip const
 *
 *   S-a adaugat:
 *     -  procedura de modificare a dimensiunii din Exercitiul 4.3.
 *     -  procedura de comparare lexicografica din Sectiunea 7.4.1
 *     -  clasa tmsort (sortarea prin interclasare) din Sectiunea 7.4.2.
 *
 *   Operatorul de indexare a fost declarat virtual pentru a
 *   permite legarea dinamica in cazul tablourilor initializate
 *   virtual (Sectiunea 10.2.2)
 *
 *    S-a introdus functia tip( ) pentru RTTI (Sectiunea 10.2.3).
 */

#ifndef __TABLOU_H
#define __TABLOU_H

#include <iostream.h>

template <class T>
class tablou {
  friend class tmsort<T>;

public:
  // constructorii si destructorul
  tablou( int = 0 );           // constructor (numarul de elemente)
  tablou( const tablou& );     // constructor de copiere
 ~tablou( )  { delete [ ] a; } // elibereaza memoria alocata

  // operatori de atribuire si indexare
  tablou& operator =( const tablou& );
  virtual T& operator []( int );
  const T& operator []( int ) const;

  // returneaza numarul elementelor
  size( ) const { return d; }

  // activarea/dezactivarea verificarii indicilor
  void vOn ( ) { v = 1; }
  void vOff( ) { v = 0; }

  // redimensionarea tabloului
  tablou& newsize( int );

  // RTTI (Sectiunea 10.2.3)
  virtual char tip( ) const { return 'T'; }

protected:
  int  d; // numarul elementelor (dimensiunea) tabloului
  T *a;   // adresa zonei alocate
  char v; // indicator verificare indice

  // functie auxiliara de initializare
  void init( const tablou& );
};

template<class T>
tablou<T>::tablou( int dim ) {
  a = 0; v = 0; d = 0;      // valori implicite
  if ( dim > 0 )            // verificarea dimensiunii
    a = new T [ d = dim ];  // alocarea memoriei
}

template <class T>
tablou<T>::tablou( const tablou<T>& t ) {
  // initializarea obiectului invocator cu t
  init( t );
}

template <class T>
tablou<T>& tablou<T>::operator =( const tablou<T>& t ) {
  if ( this != &t ) { // este o atribuire inefectiva x = x?
    delete [ ] a;     // eliberarea memoriei alocate
    init( t );        // initializarea cu t
  }
  return *this;       // se returneaza obiectul invocator
}

template<class T>
void tablou<T>::init( const tablou<T>& t ) {
  a = 0; v = 0; d = 0;     // valori implicite
  if ( t.d > 0 ) {         // verificarea dimensiunii
    a = new T [ d = t.d ]; // alocarea si copierea elem.
    for ( int i = 0; i < d; i++ ) a[ i ] = t.a[ i ];
    v = t.v;               // duplicarea indicatorului
  }                        // pentru verificarea indicilor
}

template< class T >
T& tablou<T>::operator []( int i ) {
  static T z;   // elementul returnat in caz de eroare

  if ( d == 0 ) // tablou de dimensiune zero
    return z;

  if ( v == 0 || ( 0 <= i && i < d ) )
    // verificarea indicilor este dezactivata,
    // sau este activata si indicele este corect
    return a[ i ];

  cerr << "\n\ntablou -- " << i
       << ": indice exterior domeniului [0, "
       << ( d - 1 ) << "].\n\n";

  return z;
}

template< class T >
const T& tablou<T>::operator []( int i ) const {
  static T z;   // elementul returnat in caz de eroare

  if ( d == 0 ) // tablou de dimensiune zero
    return z;

  if ( v == 0 || ( 0 <= i && i < d ) )
    // verificarea indicilor este dezactivata,
    // sau este activata si indicele este corect
    return a[ i ];

  cerr << "\n\ntablou -- " << i
       << ": indice exterior domeniului [0, "
       << ( d - 1 ) << "].\n\n";

  return z;
}

template< class T >
tablou<T>& tablou<T>::newsize( int dN ) {
  T *aN = 0;              // noua adresa

  if ( dN > 0 ) {
    aN = new T [ dN ];    // alocarea dinamica a memoriei
    for ( int i = d < dN? d: dN; i--; )
      aN[ i ] = a[ i ];   // alocarea dinamica a memoriei
  }
  else
    dN = 0;

  delete [ ] a;           // eliberarea vechiului spatiu
  d = dN; a = aN;         // redimensionarea obiectului

  return *this;
}

template <class T>
istream& operator >>( istream& is, tablou<T>& t ) {
  char c;

  // citirea dimensiunii tabloului incadrata de '[' si ']'
  is >> c;
  if ( c != '[' ) { is.clear( ios::failbit ); return is; }
  int n; is >> n; is >> c;
  if ( c != ']' ) { is.clear( ios::failbit ); return is; }

  // modificarea dimensiunii tabloului,
  // evitand copierea elementelor existente
  t.newsize( 0 ).newsize( n );

  // citirea elementelor
  for ( int i = 0; i < n; is >> t[ i++ ] );

  return is;
}

template <class T>
ostream& operator <<( ostream& os, tablou<T>& t ) {
  int n = t.size( );

  os << " [" << n << "]: ";
  for ( int i = 0; i < n; os << t[ i++ ] << ' ' );

  return os;
}


/*
 * Aceste proceduri sunt necesare datorita unor
 * anomali din bibliotecile de I/E Borland 3.1 si 4.0.
 *
 * In mod normal, argumentul tablou<T> din operatorul de insesare <<
 * ar trebui declarat const. Am renutat, deoarece si de aceasta data
 * Borland-ul s-a comportat necorespunzator.
 *
 * Banuiesc ca sunt ceva probleme cu derivarile publice ale
 * claselor parametrice.
 */
 
/*
template <class T>
istream& operator >>( istream_withassign& is, tablou<T>& t ) {
  return ( is >> dec >> t );
}

template <class T>
ostream& operator <<( ostream_withassign& os, tablou<T>& t ) {
  return ( os << dec << t );
}
 */


/*
 * sortare prin interclasare (mergesort)
 */

template <class T>
tablou<T>& mergesort( tablou<T>& t ) {
  ( tmsort<T> )t;
  return t;
}


// clasa pentru implementarea sortarii prin interclasare
template <class T>
class tmsort {
public:
  tmsort( tablou<T>& );

private:
  T *a, *x;

  void mergesort( int, int );
};

template <class T>
tmsort<T>::tmsort( tablou<T>& t ): a( t.a ) {
  x = new T[ t.size( ) ];
  mergesort( 0, t.size( ) );
  delete [ ] x;
}

template <class T>
void tmsort<T>::mergesort( int st, int dr ) {
  if ( dr - st > 1 ) {
    // mijlocul intervalului
    int m = ( st + dr ) / 2;

    // sortarea celor doua parti
    mergesort( st, m );
    mergesort( m, dr );

    // pregatirea zonei x pentru interclasare
    int k = st;
    for ( int i = st; i < m; ) x[ i++ ] = a[ k++ ];
    for ( int j = dr; j > m; ) x[ --j ] = a[ k++ ];

    // interclasarea celor doua parti din x in zona a
    i = st; j = dr - 1;
    for ( k = st; k < dr; k++ )
    a[ k ] = x[ j ] > x[ i ]? x[ i++ ]: x[ j-- ];
  }
}


/*
 * operatori de comparare pentru tablouri
 */

template <class T>
operator > ( const tablou<T>& a, const tablou<T>& b ) {
  // minumul elementelor
  int as = a.size( ), bs = b.size( );
  int n = as < bs? as: bs;

  // comparam pana la prima diferenta
  for ( int i = 0; i < n; i++ )
    if ( a[ i ] != b[ i ] ) return a[ i ] > b[ i ];

  // primele n elemente sunt identice
  return as > bs;
}

#endif

